use multi_key_manager::{file::open_rwlocked, file::File, LiveKeys};
use std::{
    fs::create_dir_all,
    path::{Path, PathBuf},
    sync::Arc,
};

pub mod errors;
use errors::{CvfsErrors, CvfsResult};
mod aead;
use aead::{decrypt_file_buffered, encrypt_file_buffered};
pub mod metadata;
use metadata::{FileInfo, Metadata};
mod secure_file;
use secure_file::SecureFileHandler;

const SUPERBLOCK_PATH: &str = "encfsstorage.superblock";
const BLOCK_PATH: &str = "blocks/";
const METADATA_PATH: &str = "metadata.index";

pub fn open_dir<P: AsRef<Path>>(
    decryption_input: &str,
    directory_path: P,
) -> Result<CryptoVFS, CvfsErrors> {
    let directory_path = directory_path.as_ref();
    create_dir_all(directory_path)?;
    let keys_path = path_extend(directory_path, SUPERBLOCK_PATH);
    if keys_path.is_dir() {
        return Err(CvfsErrors::GeneralIO(format!(
            "keys_path must not be a directory: {:?}",
            keys_path.display()
        )));
    }
    let metadata_path = path_extend(directory_path, METADATA_PATH);
    if metadata_path.is_dir() {
        return Err(CvfsErrors::GeneralIO(format!(
            "metadata_path must not be a directory: {:?}",
            metadata_path.display()
        )));
    }
    let mut metadata_file = open_rwlocked(&metadata_path)?;

    let blocks_path = path_extend(directory_path, BLOCK_PATH);
    create_dir_all(blocks_path)?;

    let mut keys = if keys_path.is_file() {
        LiveKeys::read_from_disk(decryption_input, keys_path)
    } else {
        LiveKeys::generate_new(decryption_input, keys_path)
    }?;

    let metadata = if metadata_file.metadata()?.len() == 0 {
        Metadata::empty()
    } else {
        let content = decrypt_file_buffered(&mut metadata_file, &mut keys)?;
        Metadata::restore_from_slice(&content)?
    };

    Ok(CryptoVFS {
        keys,
        metadata: Arc::new(metadata),
    })
}

/// A mounted virtual file system
pub struct CryptoVFS {
    keys: LiveKeys,
    metadata: Arc<Metadata>,
}
impl CryptoVFS {
    /// Opens a file in write-only mode.
    ///
    /// This function will create a file if it does not exist, and will truncate it if it does.
    pub fn create(&mut self, path: &str) -> CvfsResult<SecureFileHandler> {
        //let weak_metadata = Arc::downgrade(&self.metadata);
        let metadata =
            Arc::get_mut(&mut self.metadata).expect("CryptoVFS::create double get_mut(metadata)");

        let entry: &mut FileInfo = metadata
            .files
            .entry(path.into())
            .or_insert_with(|| FileInfo::new(path));
        todo!()
    }

    pub fn exists(&self, path: &str) -> bool {
        self.metadata.files.contains_key(path)
    }
}

pub fn path_extend<B: AsRef<Path>, E: AsRef<Path>>(base: B, ext: E) -> PathBuf {
    let mut buf = base.as_ref().to_owned();
    buf.push(ext);
    buf
}

#[cfg(test)]
mod tests {
    use super::*;

    /// ext must not start with a `/`
    fn path<P: AsRef<Path>>(ext: P) -> PathBuf {
        let mut p = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
        // we are in a cargo workspace
        p.pop();
        p.push("target");
        p.push("aead-vfs_tests");
        std::fs::create_dir_all(&p).expect("unable to create testing directory");
        p.push(ext);
        p
    }

    #[test]
    fn it_works() {
        let result = open_dir("bla bla secure key bla", path("it_works"));
        assert!(result.is_ok());
    }

    #[test]
    fn encrypt_decrypt_cycle() {
        let mut keys =
            LiveKeys::generate_new("paper_key paper_key", path("encrypt_decrypt_cycle.keys"))
                .unwrap();
        let mut file = open_rwlocked(path("encrypt_decrypt_cycle.bin")).unwrap();
        let random_data = b"abcdefghijklmnopqrstuvwxyz";
        println!(" >> DBG: {}", random_data.len());

        println!("nonce_before_encrypt: {:?}", keys.get_opening_nonce());
        encrypt_file_buffered(&mut file, random_data, &mut keys).expect("unable to encrypt");

        let nonce_after_encrypt = keys.get_opening_nonce();
        println!("nonce_after_encrypt: {nonce_after_encrypt:?}");

        let decrypted = decrypt_file_buffered(&mut file, &mut keys).expect("unable to decrypt");

        let nonce_after_decrypt = keys.get_opening_nonce();

        assert_eq!(random_data, &*decrypted);
    }

    /*
        #[test]
        fn encrypt_decrypt_cycle_large_file() {
            let mut keys = LiveKeys::generate_new(
                "paper_key paper_key",
                path("encrypt_decrypt_cycle_large_file.keys"),
            )
            .unwrap();
            let mut file = open_rwlocked(path("encrypt_decrypt_cycle_large_file.bin")).unwrap();
            let base_date = b"abcdefghijklmnopqrstuvwxyz";
            let mut random_data = Vec::new();
            for _ in 0..1024 * 1024 {
                random_data.extend_from_slice(base_date);
            }

            println!(" >> DBG: {}", random_data.len());

            println!("nonce_before_encrypt: {:?}", keys.get_opening_nonce());
            encrypt_file_buffered(&mut file, &random_data, &mut keys).expect("unable to encrypt");

            let nonce_after_encrypt = keys.get_opening_nonce();
            println!("nonce_after_encrypt: {nonce_after_encrypt:?}");

            let decrypted = decrypt_file_buffered(&mut file, &mut keys).expect("unable to decrypt");

            let nonce_after_decrypt = keys.get_opening_nonce();

            assert_eq!(random_data, decrypted);
        }
    // */

    #[test]
    fn double_store_and_decrypt_cycle() {
        let mut keys = LiveKeys::generate_new(
            "paper_key paper_key",
            path("double_store_and_decrypt_cycle.keys"),
        )
        .unwrap();
        let mut file = open_rwlocked(path("double_store_and_decrypt_cycle.bin")).unwrap();
        let initial_data = b"alksdjflkasjdlfkjasldkfjlaksjdfl";
        let random_data = b"abcdefghijklmnopqrstuvwxyz";
        println!(" >> DBG: {}", random_data.len());

        println!("1: {:?}", keys.get_opening_nonce());
        encrypt_file_buffered(&mut file, initial_data, &mut keys).expect("unable to encrypt");

        println!("2: {:?}", keys.get_opening_nonce());
        encrypt_file_buffered(&mut file, random_data, &mut keys).expect("unable to encrypt");

        let nonce_after_encrypt = keys.get_opening_nonce();
        println!("3: {nonce_after_encrypt:?}");

        let decrypted = decrypt_file_buffered(&mut file, &mut keys);

        let nonce_after_decrypt = keys.get_opening_nonce();
        println!("4: {nonce_after_decrypt:?}");
        let decrypted = decrypted.expect("unable to decrypt");

        assert_eq!(random_data, &*decrypted);
    }

    #[test]
    fn file_does_not_exist() {
        let vfs = open_dir("bla bla secure key bla", path("file_does_not_exist")).unwrap();
        assert!(vfs.exists("something") == false);
    }

    #[test]
    fn file_does_exist() {
        let mut vfs = open_dir("bla bla secure key bla", path("file_does_exist")).unwrap();
        let handler = vfs.create("something").unwrap();
        assert!(vfs.exists("something"));
    }
}
